/*
 * Moxa USB Serial Converter stuff
 *
 *  Copyright (C) 2019        Moxa Inc.
 *	Copyright (C) 1999 - 2012 Greg Kroah-Hartman (greg@kroah.com)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; version 2 of the License.
 *
 */

#ifndef __MX_USB_SERIAL_0_H_
#define __MX_USB_SERIAL_0_H_

#include <linux/kref.h>
#include <linux/mutex.h>
#include <linux/serial.h>
#include <linux/sysrq.h>
#include <linux/kfifo.h>

/* The maximum number of ports one device can grab at once */
#define MAX_NUM_MXPORTS		16

/**
 * mxusb_serial - structure used by the usb-serial core for a device
 * @dev: pointer to the struct usb_device for this device
 * @type: pointer to the struct usb_serial_driver for this device
 * @interface: pointer to the struct usb_interface for this device
 * @num_ports: the number of ports this device has
 * @num_interrupt_in: number of interrupt in endpoints we have
 * @num_interrupt_out: number of interrupt out endpoints we have
 * @num_bulk_in: number of bulk in endpoints we have
 * @num_bulk_out: number of bulk out endpoints we have
 * @port: array of struct usb_serial_port structures for the different ports.
 * @private: place to put any driver specific information that is needed.  The
 *	usb-serial driver is required to manage this data, the usb-serial core
 *	will not touch this.  Use usb_get_serial_data() and
 *	usb_set_serial_data() to access this.
 */
struct mxusb_serial {
	struct usb_device		*dev;
	struct usb_serial_driver	*type;
	struct usb_interface		*interface;
	unsigned char			disconnected:1;
	unsigned char			suspending:1;
	unsigned char			attached:1;
	unsigned char			minors_reserved:1;
	unsigned char			num_ports;
	unsigned char			num_port_pointers;
	char				num_interrupt_in;
	char				num_interrupt_out;
	char				num_bulk_in;
	char				num_bulk_out;
	struct usb_serial_port		*port[MAX_NUM_PORTS];
	struct kref			kref;
	struct mutex			disc_mutex;
	void				*private;
};
#define to_mxusb_serial(d) container_of(d, struct mxusb_serial, kref)

extern int mxusb_serial_register_drivers(struct usb_serial_driver *const serial_drivers[],
		const char *name, const struct usb_device_id *id_table);
extern void mxusb_serial_deregister_drivers(struct usb_serial_driver *const serial_drivers[]);
extern void mxusb_serial_port_softint(struct usb_serial_port *port);

extern int mxusb_serial_suspend(struct usb_interface *intf, pm_message_t message);
extern int mxusb_serial_resume(struct usb_interface *intf);

/* USB Serial console functions */
extern void usb_serial_console_init(int minor);
extern void usb_serial_console_exit(void);
extern void usb_serial_console_disconnect(struct usb_serial *serial);

/* Functions needed by other parts of the usbserial core */
extern struct usb_serial_port *mxusb_serial_port_get_by_minor(unsigned int minor);
extern void mxusb_serial_put(struct mxusb_serial *serial);
extern int mxusb_serial_generic_open(struct tty_struct *tty,
	struct usb_serial_port *port);
extern int mxusb_serial_generic_write_start(struct usb_serial_port *port,
							gfp_t mem_flags);
extern int mxusb_serial_generic_write(struct tty_struct *tty,
	struct usb_serial_port *port, const unsigned char *buf, int count);
extern void mxusb_serial_generic_close(struct usb_serial_port *port);
extern int mxusb_serial_generic_resume(struct usb_serial *serial);
extern int mxusb_serial_generic_write_room(struct tty_struct *tty);
extern int mxusb_serial_generic_chars_in_buffer(struct tty_struct *tty);
extern void mxusb_serial_generic_wait_until_sent(struct tty_struct *tty,
								long timeout);
extern void mxusb_serial_generic_read_bulk_callback(struct urb *urb);
extern void mxusb_serial_generic_write_bulk_callback(struct urb *urb);
extern void mxusb_serial_generic_throttle(struct tty_struct *tty);
extern void mxusb_serial_generic_unthrottle(struct tty_struct *tty);
extern int mxusb_serial_generic_tiocmiwait(struct tty_struct *tty,
							unsigned long arg);
extern int mxusb_serial_generic_get_icount(struct tty_struct *tty,
					struct serial_icounter_struct *icount);
extern int mxusb_serial_generic_register(void);
extern void mxusb_serial_generic_deregister(void);
extern int mxusb_serial_generic_submit_read_urbs(struct usb_serial_port *port,
						 gfp_t mem_flags);
extern void mxusb_serial_generic_process_read_urb(struct urb *urb);
extern int mxusb_serial_generic_prepare_write_buffer(struct usb_serial_port *port,
						void *dest, size_t size);
extern int mxusb_serial_handle_sysrq_char(struct usb_serial_port *port,
					unsigned int ch);
extern int mxusb_serial_handle_break(struct usb_serial_port *port);
extern void mxusb_serial_handle_dcd_change(struct usb_serial_port *usb_port,
					 struct tty_struct *tty,
					 unsigned int status);


extern int mxusb_serial_bus_register(struct usb_serial_driver *device);
extern void mxusb_serial_bus_deregister(struct usb_serial_driver *device);

extern struct usb_serial_driver mxusb_serial_generic_device;
extern struct bus_type mxusb_serial_bus_type;
extern struct tty_driver *mxusb_serial_tty_driver;

static inline void mxusb_serial_debug_data(struct device *dev,
					 const char *function, int size,
					 const unsigned char *data)
{
	dev_dbg(dev, "%s - length = %d, data = %*ph\n",
		function, size, size, data);
}

/*
 * Macro for reporting errors in write path to avoid inifinite loop
 * when port is used as a console.
 */
#define mxdev_err_console(usport, fmt, ...)				\
do {									\
	static bool __print_once;					\
	struct usb_serial_port *__port = (usport);			\
									\
	if (!__port->port.console || !__print_once) {			\
		__print_once = true;					\
		dev_err(&__port->dev, fmt, ##__VA_ARGS__);		\
	}								\
} while (0)

/*
 * module_usb_serial_driver() - Helper macro for registering a USB Serial driver
 * @__serial_drivers: list of usb_serial drivers to register
 * @__ids: all device ids that @__serial_drivers bind to
 *
 * Helper macro for USB serial drivers which do not do anything special
 * in module init/exit. This eliminates a lot of boilerplate. Each
 * module may only use this macro once, and calling it replaces
 * module_init() and module_exit()
 *
 */
#define mxusb_serial_module_driver(__name, __serial_drivers, __ids)	\
static int __init mxusb_serial_module_init(void)				\
{									\
	return mxusb_serial_register_drivers(__serial_drivers,		\
					   __name, __ids);		\
}									\
module_init(mxusb_serial_module_init);					\
static void __exit mxusb_serial_module_exit(void)				\
{									\
	mxusb_serial_deregister_drivers(__serial_drivers);		\
}									\
module_exit(mxusb_serial_module_exit);

#define module_mxusb_serial_driver(__serial_drivers, __ids)		\
	mxusb_serial_module_driver(KBUILD_MODNAME, __serial_drivers, __ids)

#endif /* __MX_USB_SERIAL_0_H */

